# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from collections import Callable
from enum import Enum
from typing import Dict, Any, List



from model_compression_toolkit.common.defaultdict import DefaultDict
from model_compression_toolkit.common.graph.base_node import BaseNode
from model_compression_toolkit.common.target_platform.op_quantization_config import QuantizationMethod


class ChannelAxis(Enum):
    """

    Index of output channels axis:

    NHWC - Output channels index is last.

    NCHW - Output channels index is 1.

    """
    NHWC = -1
    NCHW = 1


class FrameworkInfo(object):

    def __init__(self,
                 activation_quantizer_mapping: Dict[QuantizationMethod, Callable],
                 weights_quantizer_mapping: Dict[QuantizationMethod, Callable],
                 kernel_channels_mapping: DefaultDict,
                 activation_min_max_mapping: Dict[str, tuple],
                 layer_min_max_mapping: Dict[Any, tuple],
                 kernel_ops_attributes_mapping: DefaultDict,
                 output_channel_index: ChannelAxis):
        """
        A class to wrap all information about a specific framework the library needs to quantize a model.
        Specifically, FrameworkInfo holds lists of layers by how they should be quantized, and multiple mappings such as
        layer to it kernel channels indices, and a layer to its min/max values, etc.
        The layers lists are divided into three groups:
        kernel_ops: Layers that have coefficients and need to get quantized (e.g., Conv2D, Dense, etc.)
        activation_ops: Layers that their outputs should get quantized (e.g., Add, ReLU, etc.)
        no_quantization_ops:Layers that should not get quantized (e.g., Reshape, Transpose, etc.)

        Args:
            activation_quantizer_mapping (Dict[QuantizationMethod, Callable]): A dictionary mapping from QuantizationMethod to a quantization function.
            weights_quantizer_mapping (Dict[QuantizationMethod, Callable]): A dictionary mapping from QuantizationMethod to a quantization function.
            kernel_channels_mapping (DefaultDict): Dictionary from a layer to a tuple of its kernel in/out channels indices.
            activation_min_max_mapping (Dict[str, tuple]): Dictionary from an activation function to its min/max output values.
            layer_min_max_mapping (Dict[Any, tuple]): Dictionary from a layer to its min/max output values.
            kernel_ops_attributes_mapping (DefaultDict): Dictionary from a framework operator to a list of its weights attirbutes to quantize.
            output_channel_index (ChannelAxis): Index of output channels of the model's layers (for computing statistics per-channel).

        Examples:
            When quantizing a Keras model, if we want to quantize the kernels of Conv2D layers only, we can
            set, and we know it's kernel out/in channel indices are (3, 2) respectivly:

            >>> import tensorflow as tf
            >>> kernel_ops = [tf.keras.layers.Conv2D]
            >>> kernel_channels_mapping = DefaultDict({tf.keras.layers.Conv2D: (3,2)})

            Then, we can create a FrameworkInfo object:

            >>> FrameworkInfo(kernel_channels_mapping, {}, {})

            If an activation layer (tf.keras.layers.Activation) should be quantized and we know it's min/max outputs range in advanced, we can add it to activation_min_max_mapping for saving the statistics collection time. For example:

            >>> activation_min_max_mapping = {'softmax': (0, 1)}
            >>> FrameworkInfo(kernel_channels_mapping, activation_min_max_mapping, {})

            If a layer's activations should be quantized and we know it's min/max outputs range in advanced, we can add it to layer_min_max_mapping for saving the statistics collection time. For example:

            >>> layer_min_max_mapping = {tf.keras.layers.Softmax: (0, 1)}
            >>> FrameworkInfo(kernel_channels_mapping, activation_min_max_mapping, layer_min_max_mapping)

        """

        self.activation_quantizer_mapping = activation_quantizer_mapping
        self.weights_quantizer_mapping = weights_quantizer_mapping
        self.kernel_channels_mapping = kernel_channels_mapping
        self.activation_min_max_mapping = activation_min_max_mapping
        self.layer_min_max_mapping = layer_min_max_mapping
        self.kernel_ops_attributes_mapping = kernel_ops_attributes_mapping
        self.output_channel_index = output_channel_index


    def get_kernel_op_attributes(self, node_type: Any) -> List[str]:
        """
        Get a list of attributes of a layer's weights to quantize.

        Args:
            node_type: Layer to get its' attributes.

        Returns:
            A list of attributes the layer has and should be quantized.
        """
        attr_list = self.kernel_ops_attributes_mapping.get(node_type)
        return attr_list

    def layers_has_min_max(self, layer: Any) -> bool:
        """
        Check if a layer is in a layer to min/max mapping the FrameworkInfo holds.
        Args:
            layer: A layer to check if has a min/max known values.

        Returns:
            Whether a layer has a min/max known values or not.
        """

        return layer in self.layer_min_max_mapping

    def activation_has_min_max(self, activation_name: str) -> bool:
        """
        Check if an activation layer has a min/max mapping.

        Args:
            activation_name: String of the activation function to check for its min/max values.

        Returns:
            Whether an activation layer has a min/max known values or not.
        """

        return activation_name in self.activation_min_max_mapping



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from torch import reshape
import torch
from model_compression_toolkit.common.graph.graph_matchers import NodeOperationMatcher
from model_compression_toolkit import common
from model_compression_toolkit.common.graph.base_graph import Graph
from model_compression_toolkit.common.graph.base_node import BaseNode


class ReshapeWithStaticShapes(common.BaseSubstitution):
    """
    Replace "reshape" or "view" shape attributes. Shape attributes are replaced to static const values.
    """

    def __init__(self):
        """
        Matches: 'reshape' or 'view' operators.
        """
        nodes = NodeOperationMatcher(reshape) | NodeOperationMatcher(torch.Tensor.view)
        super().__init__(matcher_instance=nodes)

    def substitute(self,
                   graph: Graph,
                   node: BaseNode) -> Graph:
        """
        Replaces the 'size' attribute to 'reshape' or 'view' operators to be a list of integers,
        determined by their intended output shape. This replaces 'size' attributes that come from
        nodes in the graph. We delete nodes for which that was their sole purpose.

        Args:
            graph: Graph we apply the substitution on.
            node: node that match the pattern in the substitution init.

        Returns:
            Graph after applying the substitution.
        """
        # configure the new static output shape attribute
        node.op_call_args = node.output_shape

        # modify the node input info
        node.input_shape = [node.input_shape[0]]
        nodes_to_check = []
        for in_edge in graph.incoming_edges(node):
            if in_edge.sink_index > 0: # the first input is the tensor to be reshaped
                nodes_to_check.append(in_edge.source_node)
                graph.remove_edge(in_edge.source_node, node)
        for n in nodes_to_check:
            clean_graph_from_nodes_without_out_edges(graph, n)
        return graph


def clean_graph_from_nodes_without_out_edges(graph: Graph,
                                             node: BaseNode):
    """
    This function removes all the nodes and edges to nodes that are not connected to any other node,
    and are not output nodes.
    Args:
        graph: Graph we apply the substitution on.
        node: node that match the pattern in the substitution init.

    """
    output_nodes = [o.node for o in graph.get_outputs()]
    if len(graph.out_edges(node)) == 0 and node not in output_nodes:
        nodes_to_check = []
        for in_edge in graph.incoming_edges(node):
            nodes_to_check.append(in_edge.source_node)
            graph.remove_edge(in_edge.source_node, node)
        graph.remove_node(node)
        for n in nodes_to_check:
            clean_graph_from_nodes_without_out_edges(graph, n)


# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import copy
from enum import Enum
from typing import List

class QuantizationMethod(Enum):
    """
    Method for quantization function selection:

    POWER_OF_TWO - Symmetric, uniform, threshold is power of two quantization.

    KMEANS - k-means quantization.

    LUT_QUANTIZER - quantization using a look up table.

    SYMMETRIC - Symmetric, uniform, quantization.

    UNIFORM - uniform quantization,

    """
    POWER_OF_TWO = 0
    KMEANS = 1
    LUT_QUANTIZER = 2
    SYMMETRIC = 3
    UNIFORM = 4




class OpQuantizationConfig:
    """
    OpQuantizationConfig is a class to configure the quantization parameters of an operator.
    """

    def __init__(self,
                 activation_quantization_method: QuantizationMethod,
                 weights_quantization_method: QuantizationMethod,
                 activation_n_bits: int,
                 weights_n_bits: int,
                 weights_per_channel_threshold: bool,
                 enable_weights_quantization: bool,
                 enable_activation_quantization: bool,
                 quantization_preserving: bool,
                 fixed_scale: float,
                 fixed_zero_point: int,
                 weights_multiplier_nbits: int  # If None - set 8 in hptq, o.w use it
                 ):
        """

        Args:
            activation_quantization_method (QuantizationMethod): Which method to use from QuantizationMethod for activation quantization.
            weights_quantization_method (QuantizationMethod): Which method to use from QuantizationMethod for weights quantization.
            activation_n_bits (int): Number of bits to quantize the activations.
            weights_n_bits (int): Number of bits to quantize the coefficients.
            weights_per_channel_threshold (bool): Whether to quantize the weights per-channel or not (per-tensor).
            enable_weights_quantization (bool): Whether to quantize the model weights or not.
            enable_activation_quantization (bool): Whether to quantize the model activations or not.
            quantization_preserving (bool): Whether quantization parameters should be the same for an operator's input and output.
            fixed_scale (float): Scale to use for an operator quantization parameters.
            fixed_zero_point (int): Zero-point to use for an operator quantization parameters.
            weights_multiplier_nbits (int): Number of bits to use when quantizing in look-up-table.
        """

        self.activation_quantization_method = activation_quantization_method
        self.weights_quantization_method = weights_quantization_method
        self.activation_n_bits = activation_n_bits
        self.weights_n_bits = weights_n_bits
        self.weights_per_channel_threshold = weights_per_channel_threshold
        self.enable_weights_quantization = enable_weights_quantization
        self.enable_activation_quantization = enable_activation_quantization
        self.quantization_preserving = quantization_preserving
        self.fixed_scale = fixed_scale
        self.fixed_zero_point = fixed_zero_point
        self.weights_multiplier_nbits = weights_multiplier_nbits

    def get_info(self):
        """

        Returns: Info about the quantization configuration as a dictionary.

        """
        return self.__dict__

    def clone_and_edit(self, **kwargs):
        """
        Clone the quantization config and edit some of its attributes.
        Args:
            **kwargs: Keyword arguments to edit the configuration to clone.

        Returns:
            Edited quantization configuration.
        """

        qc = copy.deepcopy(self)
        for k, v in kwargs.items():
            assert hasattr(qc,
                           k), f'Edit attributes is possible only for existing attributes in configuration, ' \
                               f'but {k} is not an attribute of {qc}'
            setattr(qc, k, v)
        return qc

    def __eq__(self, other):
        """
        Is this configuration equal to another object.
        Args:
            other: Object to compare.

        Returns:
            Whether this configuration is equal to another object or not.
        """
        if not isinstance(other, OpQuantizationConfig):
            return False
        return self.activation_quantization_method == other.activation_quantization_method and \
               self.weights_quantization_method == other.weights_quantization_method and \
               self.activation_n_bits == other.activation_n_bits and \
               self.weights_n_bits == other.weights_n_bits and \
               self.weights_per_channel_threshold == other.weights_per_channel_threshold and \
               self.enable_weights_quantization == other.enable_weights_quantization and \
               self.enable_activation_quantization == other.enable_activation_quantization


class QuantizationConfigOptions(object):
    """

    Wrap a set of quantization configurations to consider during the quantization
    of an operator.

    """
    def __init__(self,
                 quantization_config_list: List[OpQuantizationConfig],
                 base_config: OpQuantizationConfig = None):
        """

        Args:
            quantization_config_list (List[OpQuantizationConfig]): List of possible OpQuantizationConfig to gather.
            base_config (OpQuantizationConfig): Fallback OpQuantizationConfig to use when optimizing the model in a non mixed-precision manner.
        """

        assert isinstance(quantization_config_list,
                          list), f'QuantizationConfigOptions options list should be of type list, but is: ' \
                                 f'{type(quantization_config_list)}'
        assert len(quantization_config_list) > 0, f'Options list can not be empty'
        for cfg in quantization_config_list:
            assert isinstance(cfg, OpQuantizationConfig), f'Options should be a list of QuantizationConfig objects, ' \
                                                        f'but found an object type: {type(cfg)}'
        self.quantization_config_list = quantization_config_list
        if len(quantization_config_list) > 1:
            assert base_config is not None, f'When quantization config options contains more than one configuration, ' \
                                            f'a base_config must be passed for non-mixed-precision optimization process'
            assert base_config in quantization_config_list, f"base_config must be in the given quantization config " \
                                                            f"list of options"
            self.base_config = base_config
        elif len(quantization_config_list) == 1:
            self.base_config = quantization_config_list[0]
        else:
            raise Exception("QuantizationConfigOptions must have at least one OpQuantizationConfig "
                            "defined in its options list, but list is empty")

    def __eq__(self, other):
        """
        Is this QCOptions equal to another object.
        Args:
            other: Object to compare.

        Returns:
            Whether this QCOptions equal to another object or not.
        """

        if not isinstance(other, QuantizationConfigOptions):
            return False
        if len(self.quantization_config_list) != len(other.quantization_config_list):
            return False
        for qc, other_qc in zip(self.quantization_config_list, other.quantization_config_list):
            if qc != other_qc:
                return False
        return True

    def clone_and_edit(self, **kwargs):
        qc_options = copy.deepcopy(self)
        for qc in qc_options.quantization_config_list:
            self.__edit_quantization_configuration(qc, kwargs)
        return qc_options

    def __edit_quantization_configuration(self, qc, kwargs):
        for k, v in kwargs.items():
            assert hasattr(qc,
                           k), f'Edit attributes is possible only for existing attributes in configuration, ' \
                               f'but {k} is not an attribute of {qc}'
            setattr(qc, k, v)

    def get_info(self):
        return {f'option {i}': cfg.get_info() for i, cfg in enumerate(self.quantization_config_list)}




# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

# Supported frameworks in MCT:
TENSORFLOW = 'tensorflow'
PYTORCH = 'pytorch'

# Minimal threshold to use for quantization ranges:
MIN_THRESHOLD = (2 ** -28)
EPS = 1e-8
MULTIPLIER_N_BITS = 8

# Quantization attributes:
OUTPUT_SCALE = 'output_scale'
THRESHOLD = 'threshold'
SIGNED = 'is_signed'
CLUSTER_CENTERS = 'cluster_centers'
SCALE_PER_CHANNEL = 'scale_per_channel'
RANGE_MIN = 'range_min'
RANGE_MAX = 'range_max'

# BaseNode attributes
REUSE = 'reuse'
REUSE_GROUP = 'reuse_group'

# Data types:
DATA_TYPE = 'dtype'
FLOAT_32 = 'float32'


# Number of Tensorboard cosine-similarity plots to add:
NUM_SAMPLES_CS_TENSORBOARD = 20

# num bits for shift negative non linear node
SHIFT_NEGATIVE_NON_LINEAR_NUM_BITS = 16

# Default bitwidth for disabled quantization candidate
DEFAULT_CANDIDATE_BITWIDTH = 16

# In Mixed-Precision, a node can have multiple candidates for weights and activations quantization configuration.
# In order to display a single view of a node (for example, for logging in TensorBoard) we need to track the attributes
# that are shared among different candidates:
WEIGHTS_NBITS_ATTRIBUTE = 'weights_n_bits'
CORRECTED_BIAS_ATTRIBUTE = 'corrected_bias'
ACTIVATION_NBITS_ATTRIBUTE = 'activation_n_bits'


# Quantization Parameters Iterative Search Defaults:
SYMMETRIC_TENSOR_N_ITER = 40
SYMMETRIC_TENSOR_PER_CHANNEL_N_ITER = 15
SYMMETRIC_HISTOGRAM_N_ITER = 20

UNIFORM_TENSOR_N_ITER = 30
UNIFORM_TENSOR_PER_CHANNEL_N_ITER = 10
UNIFORM_HISTOGRAM_N_ITER = 30

SYMMETRIC_TENSOR_N_INTERVALS = 30
SYMMETRIC_TENSOR_PER_CHANNEL_N_INTERVALS = 30
SYMMETRIC_HISTOGRAM_N_INTERVALS = 30

SYMMETRIC_TENSOR_DEC_FREQ = 5
SYMMETRIC_TENSOR_PER_CHANNEL_DEC_FREQ = 3
SYMMETRIC_HISTOGRAM_DEC_FREQ = 4

UNIFORM_TENSOR_N_SAMPLES = 8
UNIFORM_HISTOGRAM_N_SAMPLES = 12

DEFAULT_DEC_FACTOR = (1.02, 0.98)
DEFAULT_TOL = 1e-11
BOTTOM_FACTOR = 0.7
UPPER_FACTOR = 1.2
DEC_RANGE_BOTTOM = 0.97
DEC_RANGE_UPPER = 1.03

# KPI computation parameters
BITS_TO_BYTES = 8.0



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import tensorflow as tf

from model_compression_toolkit.common.target_platform import TargetPlatformModel

if tf.__version__ < "2.6":
    from tensorflow.keras.layers import Conv2D, DepthwiseConv2D, Dense, Conv2DTranspose, Reshape, ZeroPadding2D, Dropout, \
        MaxPooling2D, Activation, ReLU, Add, PReLU, Flatten, Cropping2D
else:
    from keras.layers import Conv2D, DepthwiseConv2D, Dense, Conv2DTranspose, Reshape, ZeroPadding2D, \
    Dropout, MaxPooling2D, Activation, ReLU, Add, PReLU, Flatten, Cropping2D

from model_compression_toolkit.tpc_models.default_tp_model import get_default_tp_model

import model_compression_toolkit as mct
tpc = mct.target_platform


def get_default_keras_tpc():
    default_tp_model = get_default_tp_model()
    return generate_keras_default_tpc(name='default_keras_tpc',
                                      tp_model=default_tp_model)


def generate_keras_default_tpc(name: str, tp_model: TargetPlatformModel):
    """
    Generates a TargetPlatformCapabilities object with default operation sets to layers mapping.

    Args:
        name: Name of the TargetPlatformCapabilities.
        tp_model: TargetPlatformModel object.

    Returns: a TargetPlatformCapabilities object for the given TargetPlatformModel.
    """

    keras_tpc = tpc.TargetPlatformCapabilities(tp_model,
                                                name=name)
    with keras_tpc:
        tpc.OperationsSetToLayers("NoQuantization", [Reshape,
                                                     tf.reshape,
                                                     Flatten,
                                                     Cropping2D,
                                                     ZeroPadding2D,
                                                     Dropout,
                                                     MaxPooling2D,
                                                     tf.split,
                                                     tf.quantization.fake_quant_with_min_max_vars])

        tpc.OperationsSetToLayers("Conv", [Conv2D,
                                           DepthwiseConv2D,
                                           tf.nn.conv2d,
                                           tf.nn.depthwise_conv2d])

        tpc.OperationsSetToLayers("FullyConnected", [Dense])

        tpc.OperationsSetToLayers("ConvTranspose", [Conv2DTranspose,
                                                    tf.nn.conv2d_transpose])

        tpc.OperationsSetToLayers("AnyReLU", [tf.nn.relu,
                                              tf.nn.relu6,
                                              tpc.LayerFilterParams(ReLU, negative_slope=0.0),
                                              tpc.LayerFilterParams(Activation, activation="relu")])

        tpc.OperationsSetToLayers("Add", [tf.add,
                                          Add])

        tpc.OperationsSetToLayers("PReLU", [PReLU])

        tpc.OperationsSetToLayers("Swish", [tf.nn.swish,
                                            tpc.LayerFilterParams(Activation, activation="swish")])

        tpc.OperationsSetToLayers("Sigmoid", [tf.nn.sigmoid,
                                              tpc.LayerFilterParams(Activation, activation="sigmoid")])

        tpc.OperationsSetToLayers("Tanh", [tf.nn.tanh,
                                           tpc.LayerFilterParams(Activation, activation="tanh")])
    return keras_tpc



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import List

from model_compression_toolkit.common.framework_implementation import FrameworkImplementation
from model_compression_toolkit.common.framework_info import FrameworkInfo
from model_compression_toolkit.common import Graph, BaseNode, Logger
from model_compression_toolkit.common.quantization.quantization_params_generation.qparams_activations_computation \
    import \
    get_activations_qparams
from model_compression_toolkit.common.quantization.quantization_params_generation.qparams_weights_computation import \
    get_weights_qparams, get_channels_axis


def calculate_quantization_params(graph: Graph,
                                  fw_info: FrameworkInfo,
                                  nodes: List[BaseNode] = [],
                                  specific_nodes: bool = False,
                                  fw_impl: FrameworkImplementation = None):
    """
    For a graph, go over its nodes, compute quantization params (for both weights and activations according
    to the given framework info), and create and attach a NodeQuantizationConfig to each node (containing the
    computed params).
    By default, the function goes over all nodes in the graph. However, the specific_nodes flag enables
    to compute quantization paramss for specific nodes if the default behavior is unnecessary. For that,
    a list of nodes nodes should be passed as well.

    Args:
        fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices,
        groups of layers by how they should be quantized, etc.)
        graph: Graph to compute its nodes' thresholds.
        nodes: List of nodes to compute their thresholds instead of computing it for all nodes in the graph.
        specific_nodes: Flag to compute thresholds for only specific nodes.
        fw_impl: FrameworkImplementation with specific framework implementations.

    """

    # Create a list of nodes to compute their thresholds
    nodes_list: List[BaseNode] = nodes if specific_nodes else graph.nodes()

    for n in nodes_list:  # iterate only nodes that we should compute their thresholds
        for candidate_qc in n.candidates_quantization_cfg:
            if n.is_weights_quantization_enabled():
                # If node's weights should be quantized, we compute its weights' quantization parameters
                output_channels_axis, _ = get_channels_axis(candidate_qc.weights_quantization_cfg, fw_info, n.type)
                weights_params = get_weights_qparams(n.get_weights_by_keys(fw_impl.constants.KERNEL),
                                                     candidate_qc.weights_quantization_cfg,
                                                     output_channels_axis)
                candidate_qc.weights_quantization_cfg.set_weights_quantization_param(weights_params)
                candidate_qc.weights_quantization_cfg.weights_channels_axis = output_channels_axis
            if n.is_activation_quantization_enabled():
                # If node's activations should be quantized as well, we compute its activation quantization parameters
                activation_params = get_activations_qparams(
                    activation_quant_cfg=candidate_qc.activation_quantization_cfg,
                    nodes_prior_info=n.prior_info,
                    out_stats_container=graph.get_out_stats_collector(n))
                # Create a NodeQuantizationConfig containing all quantization params and attach it to the node
                candidate_qc.activation_quantization_cfg.set_activation_quantization_param(activation_params)



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from sklearn.cluster import KMeans
import numpy as np

from model_compression_toolkit.common.constants import CLUSTER_CENTERS, MIN_THRESHOLD, SCALE_PER_CHANNEL, \
    MULTIPLIER_N_BITS
from model_compression_toolkit.common.quantization.quantizers.quantizers_helpers import kmeans_assign_clusters, \
    int_quantization_with_scale, get_quantized_tensor


def lut_kmeans_quantizer(tensor_data: np.ndarray,
                        n_bits: int,
                        signed: bool,
                        quantization_params: dict,
                        per_channel: bool,
                        output_channels_axis: int) -> np.ndarray:
    """
    Quantize a tensor with given cluster centers and scale-per-channel vector.
    1. We divide tensor_data with the scale vector per channel.
    2. We scale the result to the range [-2^(MULTIPLIER_N_BITS-1), 2^(MULTIPLIER_N_BITS-1)-1].
    3. We assign cluster centers to every value, multiply by scale_per_channel and divide by 2^(MULTIPLIER_N_BITS-1).
    The result is the quantized tensor.


    Args:
        tensor_data: Tensor values to quantize.
        n_bits: Number of bits to quantize the tensor.
        signed: Whether the tensor contains negative values or not.
        quantization_params: Dictionary of specific parameters for this quantization function.
        per_channel: Whether to use separate quantization per output channel.
        output_channels_axis: Axis of the output channel.

    Returns:
        Quantized data.
    """
    cluster_centers = quantization_params[CLUSTER_CENTERS]
    scales_per_channel = quantization_params[SCALE_PER_CHANNEL]
    tensor = int_quantization_with_scale(tensor_data, scales_per_channel, MULTIPLIER_N_BITS)
    shape_before_kmeans = tensor.shape
    cluster_assignments = kmeans_assign_clusters(cluster_centers, tensor.reshape(-1, 1))
    quant_tensor = get_quantized_tensor(cluster_centers[cluster_assignments].reshape(shape_before_kmeans),
                                        scales_per_channel,
                                        MULTIPLIER_N_BITS)
    return quant_tensor



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import torch
import numpy as np

from model_compression_toolkit.pytorch.constants import CUDA, CPU


def get_working_device() -> str:
    """
    Get the working device of the environment

    Returns:
        Device "cuda" if GPU is available, else "cpu"

    """
    return torch.device(CUDA if torch.cuda.is_available() else CPU)


def set_model(model: torch.nn.Module):
    """
    Set model to work in eval mode and GPU mode if GPU is available

    Args:
        model: Pytorch model

    Returns:

    """
    model.eval()
    if torch.cuda.is_available():
        model.cuda()
    else:
        model.cpu()


def to_torch_tensor(tensor):
    """
    Convert a Numpy array to a Torch tensor.
    Args:
        tensor: Numpy array.

    Returns:
        Torch tensor converted from the input Numpy array.
    """
    working_device = get_working_device()
    if isinstance(tensor, torch.Tensor):
        return tensor.to(working_device)
    elif isinstance(tensor, list):
        return [to_torch_tensor(t) for t in tensor]
    elif isinstance(tensor, tuple):
        return (to_torch_tensor(t) for t in tensor)
    elif isinstance(tensor, np.ndarray):
        return torch.from_numpy(tensor.astype(np.float32)).to(working_device)
    else:
        raise Exception(f'Conversion of type {type(tensor)} to {type(torch.Tensor)} is not supported')


def torch_tensor_to_numpy(tensor: torch.Tensor) -> np.ndarray:
    """
    Convert a Pytorch tensor to a Numpy array.
    Args:
        tensor: Pytorch tensor.

    Returns:
        Numpy array converted from the input tensor.
    """
    if isinstance(tensor, np.ndarray):
        return tensor
    elif isinstance(tensor, list):
        return [torch_tensor_to_numpy(t) for t in tensor]
    elif isinstance(tensor, tuple):
        return tuple([torch_tensor_to_numpy(t) for t in tensor])
    elif isinstance(tensor, torch.Tensor):
        return tensor.cpu().detach().contiguous().numpy()
    else:
        raise Exception(f'Conversion of type {type(tensor)} to {type(np.ndarray)} is not supported')



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from abc import ABC
from typing import List, Any

from model_compression_toolkit.common.graph.base_node import BaseNode
from model_compression_toolkit.common.matchers import node_matcher, base_graph_filter, edge_matcher
from model_compression_toolkit.common.matchers.walk_matcher import WalkMatcherList


class GraphSearches(base_graph_filter.BaseGraphFilter, ABC):
    """
    Apply searches on graphs.
    The graph needs to have 'nodes' and 'edges' attributes, and a 'get_next_nodes' method.
    """

    def _node_filter(self, node_matcher: node_matcher.BaseNodeMatcher) -> list:
        """
        Iterate over nodes and returns the nodes in the graph that matches the matcher object.

        Args:
            node_matcher: Matcher object to apply on nodes in the graph.

        Returns:
            List of nodes that match the node_matcher.
        """

        return [n for n in self.nodes if node_matcher.apply(n)]

    def _edge_filter(self, edge_matcher: edge_matcher.BaseEdgeMatcher) -> list:
        """
        Iterate over edges and returns the edges in the graph that matches
        the edge_matcher object.

        Args:
            edge_matcher: Matcher object to apply on edge.

        Returns:
            List of edges that match.
        """

        edge_list = []
        for e in self.edges:
            if edge_matcher.apply(e) and len(self.edges(e[0])):
                edge_list.append(e)

        return edge_list

    def _walk_filter(self, walk_matcher: WalkMatcherList) -> List[BaseNode]:
        """
        Search for a list of nodes which match the list in walk_matcher.
        If one the nodes in the list (that was found in the graph) has more than one output,
        this list is not returned.

        Args:
            walk_matcher: WalkMatcherList with a list of nodes to match.

        Returns:
            A list of nodes which match the list in walk_matcher.
        """

        def walk_match(node: BaseNode,
                       node_list: List[BaseNode],
                       index: int,
                       node_matcher_list: list) -> Any:
            """
            Iterate the nodes in the graph starting from 'node', and search for its next
            nodes that matches the node next to 'node' in the node_matcher_list.
            If such a node is found, keep searching from that node, by calling walk_match recursively.

            Args:
                node: Node to check if it and its next nodes matches the list in node_matcher_list.
                node_list: A list of nodes that were found so far.
                index: Index of 'node' in the list on nodes it searches.
                node_matcher_list: List of nodes to search for.

            Returns:
                The list on nodes, if found. Otherwise, None.
            """
            if node_matcher_list[index].apply(node):
                node_list.append(node)
                if (index + 1) == len(node_matcher_list):
                    return [node_list]
                result_list = [
                    walk_match(nn, node_list.copy(), index + 1, node_matcher_list) for
                    nn in self.get_next_nodes(node) if
                    # Exclude patterns with an intermediate node with multiple outputs. If it's the last
                    # node in the matcher list, it is a valid pattern and should be checked.
                    len(self.get_next_nodes(nn)) == 1 or (index + 2) == len(node_matcher_list)]
                result_filter = [r for r_list in result_list if r_list is not None for r in r_list if
                                 r is not None and len(r) == len(node_matcher_list)]
                if len(result_filter) == 1:
                    return result_filter
                elif len(result_filter) == 0:
                    return None
                else:  # not supported yet
                    return result_filter
            else:
                return None

        matcher_list = walk_matcher.matcher_list if isinstance(walk_matcher, WalkMatcherList) else [
            walk_matcher]
        result = []

        # Walk the entire graph, node by node
        result_match_list = [walk_match(n, [], 0, matcher_list) for n in self.nodes if len(self.get_next_nodes(n)) == 1]
        # Flatten lists
        result.extend([r for r_list in result_match_list if r_list is not None for r in r_list])
        return result



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import List
import numpy as np

from torch.nn import Conv2d, ReLU6, Linear, ConvTranspose2d, Hardtanh
from torch.nn.functional import relu6, hardtanh

from model_compression_toolkit import common
from model_compression_toolkit.common.graph.base_graph import Graph
from model_compression_toolkit.common.graph.graph_matchers import NodeOperationMatcher, WalkMatcher
from model_compression_toolkit.common.graph.base_node import BaseNode
from model_compression_toolkit.pytorch.constants import KERNEL, BIAS, INPLACE, HARDTANH_MIN_VAL, HARDTANH_MAX_VAL, \
    RELU_POT_BOUND


class ReLUBoundToPowerOfTwo(common.BaseSubstitution):
    """
    Substitution to scale the weights of two linear nodes, and move the bound of non-linear between them
    (if bounded) in order to use the entire constrained range when activations are quantized.
    """

    def __init__(self):
        """
        Initialize a ReLUBoundToPowerOfTwo object.
        """
        homogeneous_activation_nodes = NodeOperationMatcher(ReLU6) | \
                                       NodeOperationMatcher(relu6) | \
                                       NodeOperationMatcher(Hardtanh) | \
                                       NodeOperationMatcher(hardtanh)

        op2d_node = NodeOperationMatcher(Conv2d) | \
                    NodeOperationMatcher(Linear) | \
                    NodeOperationMatcher(ConvTranspose2d)

        wm = WalkMatcher([op2d_node,
                          homogeneous_activation_nodes,
                          op2d_node])

        self.threshold = RELU_POT_BOUND
        super().__init__(matcher_instance=wm)

    def substitute(self,
                   graph: Graph,
                   nodes_list: List[BaseNode]) -> Graph:
        """
        Transform a list of nodes in a graph to use the entire constrained quantized range.
        This is done by scaling two linear nodes with a non-linearity between them, if the non-linearity
        keeps: f(ax) = af(x).
        If the non-linearity is bounded (as ReLU6), the bound needs to be scaled as well.

        Args:
            graph: Graph to apply the substitution on.
            nodes_list: List of nodes that match the pattern in the substitution init.

        Returns:
            Graph after substitution.
        """

        first_op2d_node = nodes_list[0]
        non_linear_node = nodes_list[1]
        second_op2d_node = nodes_list[2]

        # only act on bound relu with not POT max value and 0 min value
        if non_linear_node.type == ReLU6:
            scale_factor = 6.0 / self.threshold
            non_linear_node.layer_class = Hardtanh
            non_linear_node.framework_attr[INPLACE] = False
            non_linear_node.framework_attr[HARDTANH_MIN_VAL] = 0.0
            non_linear_node.framework_attr[HARDTANH_MAX_VAL] = self.threshold
        elif non_linear_node.type == relu6:
            scale_factor = 6.0 / self.threshold
            non_linear_node.functional_op = hardtanh
            non_linear_node.functional_op.__defaults__ = (0.0, self.threshold, False)
        elif non_linear_node.type == Hardtanh:
            if (non_linear_node.framework_attr[HARDTANH_MIN_VAL] == 0.0) and not \
                    (np.log2(non_linear_node.framework_attr[HARDTANH_MAX_VAL]).astype(int) -
                     np.log2(non_linear_node.framework_attr[HARDTANH_MAX_VAL]) == 0):
                scale_factor = non_linear_node.framework_attr[HARDTANH_MAX_VAL] / self.threshold
                non_linear_node.framework_attr[HARDTANH_MAX_VAL] = self.threshold
            else:
                return graph
        elif non_linear_node.type == hardtanh:
            if (non_linear_node.framework_attr[HARDTANH_MIN_VAL] == 0.0) and not \
                    (np.log2(non_linear_node.framework_attr[HARDTANH_MAX_VAL]).astype(int) -
                     np.log2(non_linear_node.framework_attr[HARDTANH_MAX_VAL]) == 0):
                scale_factor = non_linear_node.framework_attr[HARDTANH_MAX_VAL] / self.threshold
                non_linear_node.functional_op.__defaults__ = (0.0, self.threshold, non_linear_node.framework_attr[INPLACE])
            else:
                return graph
        else:
            common.Logger.error(f"In substitution with wrong matched pattern")
        common.Logger.debug(
            f"Node named:{non_linear_node.name} changed "
            f"to:{non_linear_node.type}")

        w2_fixed = scale_factor * second_op2d_node.get_weights_by_keys(KERNEL)
        w1_fixed = first_op2d_node.get_weights_by_keys(KERNEL) / scale_factor
        b1_fixed = first_op2d_node.get_weights_by_keys(BIAS) / scale_factor

        first_op2d_node.set_weights_by_keys(KERNEL, w1_fixed)
        first_op2d_node.set_weights_by_keys(BIAS, b1_fixed)
        second_op2d_node.set_weights_by_keys(KERNEL, w2_fixed)
        return graph



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from model_compression_toolkit.common.target_platform.targetplatform2framework.current_tpc import get_current_tpc
from model_compression_toolkit.common.target_platform.targetplatform2framework.target_platform_capabilities import TargetPlatformCapabilities
from model_compression_toolkit.common.target_platform.targetplatform2framework.attribute_filter import \
    Eq, GreaterEq, NotEq, SmallerEq, Greater, Smaller
from model_compression_toolkit.common.target_platform.targetplatform2framework.layer_filter_params import \
    LayerFilterParams
from model_compression_toolkit.common.target_platform.targetplatform2framework.operations_to_layers import \
    OperationsToLayers, OperationsSetToLayers








# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import math
from enum import Enum

from model_compression_toolkit.common.constants import MIN_THRESHOLD


class QuantizationErrorMethod(Enum):
    """
    Method for quantization threshold selection:

    NOCLIPPING - Use min/max values as thresholds.

    MSE - Use min square error for minimizing quantization noise.

    MAE - Use min absolute error for minimizing quantization noise.

    KL - Use KL-divergence to make signals distributions to be similar as possible.

    Lp - Use Lp-norm to minimizing quantization noise.

    """

    NOCLIPPING = 0
    MSE = 1
    MAE = 2
    KL = 4
    LP = 5



class QuantizationConfig(object):

    def __init__(self,
                 activation_error_method: QuantizationErrorMethod = QuantizationErrorMethod.MSE,
                 weights_error_method: QuantizationErrorMethod = QuantizationErrorMethod.MSE,
                 relu_bound_to_power_of_2: bool = False,
                 weights_bias_correction: bool = True,
                 weights_per_channel_threshold: bool = True,
                 input_scaling: bool = False,
                 softmax_shift: bool = False,
                 shift_negative_activation_correction: bool = False,
                 activation_channel_equalization: bool = False,
                 z_threshold: float = math.inf,
                 min_threshold: float = MIN_THRESHOLD,
                 l_p_value: int = 2,
                 shift_negative_ratio: float = 0.05,
                 shift_negative_threshold_recalculation: bool = False):
        """
        Class to wrap all different parameters the library quantize the input model according to.

        Args:
            activation_error_method (QuantizationErrorMethod): Which method to use from QuantizationErrorMethod for activation quantization threshold selection.
            weights_error_method (QuantizationErrorMethod): Which method to use from QuantizationErrorMethod for activation quantization threshold selection.
            relu_bound_to_power_of_2 (bool): Whether to use relu to power of 2 scaling correction or not.
            weights_bias_correction (bool): Whether to use weights bias correction or not.
            weights_per_channel_threshold (bool): Whether to quantize the weights per-channel or not (per-tensor).
            input_scaling (bool): Whether to use input scaling or not.
            softmax_shift (bool): Whether to use softmax shift or not.
            shift_negative_activation_correction (bool): Whether to use shifting negative activation correction or not.
            activation_channel_equalization (bool): Whether to use activation channel equalization correction or not.
            z_threshold (float): Value of z score for outliers removal.
            min_threshold (float): Minimum threshold to use during thresholds selection.
            l_p_value (int): The p value of L_p norm threshold selection.
            shift_negative_ratio (float): Value for the ratio between the minimal negative value of a non-linearity output to its activation threshold, which above it - shifting negative activation should occur if enabled.
            shift_negative_threshold_recalculation (bool): Whether or not to recompute the threshold after shifting negative activation.

        Examples:
            One may create a quantization configuration to quantize a model according to.
            For example, to quantize a model's weights and activation using symetric uniform quantization method,
            weights threshold selection using MSE, activation threshold selection using NOCLIPPING,
            enabling relu_bound_to_power_of_2, weights_bias_correction, and quantizing the weights per-channel,
            one can instantiate a quantization configuration:

            >>> qc = QuantizationConfig(activation_error_method=QuantizationErrorMethod.NOCLIPPING,weights_error_method=QuantizationErrorMethod.MSE,relu_bound_to_power_of_2=True,weights_bias_correction=True,weights_per_channel_threshold=True)


            The QuantizationConfig instanse can then be passed to
            :func:`~model_compression_toolkit.keras_post_training_quantization`

        """

        self.activation_error_method = activation_error_method
        self.weights_error_method = weights_error_method
        self.relu_bound_to_power_of_2 = relu_bound_to_power_of_2
        self.weights_bias_correction = weights_bias_correction
        self.weights_per_channel_threshold = weights_per_channel_threshold
        self.activation_channel_equalization = activation_channel_equalization
        self.input_scaling = input_scaling
        self.softmax_shift = softmax_shift
        self.min_threshold = min_threshold
        self.shift_negative_activation_correction = shift_negative_activation_correction
        self.z_threshold = z_threshold
        self.l_p_value = l_p_value
        self.shift_negative_ratio = shift_negative_ratio
        self.shift_negative_threshold_recalculation = shift_negative_threshold_recalculation

    def __repr__(self):
        return str(self.__dict__)


# Default quantization configuration the library use.
DEFAULTCONFIG = QuantizationConfig(QuantizationErrorMethod.MSE,
                                   QuantizationErrorMethod.MSE,
                                   relu_bound_to_power_of_2=False,
                                   weights_bias_correction=True,
                                   weights_per_channel_threshold=True,
                                   input_scaling=False,
                                   softmax_shift=False)

